/* $Id: phone.c,v 1.7 2000/12/18 11:06:38 selinger Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"   /* read installation specific configuration options */

/* ---------------------------------------------------------------------- */
/** global parameters, set by main from command line and environment */

char *dialchars = DIALCHARS;    /* valid digits and chars in phone number */
int verbose = 0;                /* verbose mode? */
char *localestring = NULL;      /* phone # prefix replacement table */
int stripflag = 0;                  /* strip output? */

/* ---------------------------------------------------------------------- */
/** checks whether c is in global string dialchars */

int is_dialchar(char c) {
  return strchr(dialchars, c) != NULL;
}

/* ---------------------------------------------------------------------- */
/** Let us say strings p and q _match_ if p and q are equal after
   removing all characters that are not dialchars. The function
   match_prefix returns the length of the shortest prefix of q that
   matches p, if any, and -1 otherwise. */

int match_prefix(char *p, char *q) {
  char* q0 = q;

  while (*p != '\0') {
    if (!is_dialchar(*p)) {        /* skip non-dialchars in p */
      p++;
    } else if (!is_dialchar(*q)) { /* and _then_ in q */
      q++;
    } else if (*p == *q) {         /* do characters match? */
      p++; q++;                    
    } else {                       /* failure */
      return -1;
    }
  }    
  return q-q0;                     /* success */
}

/* ---------------------------------------------------------------------- */
/* localereader: reads a sequence of colon-separated rules of the form
   'lhs=rhs' from localestring, one for each call. Localestring is
   only passed on the first call; subsequent calls pass NULL as the
   first argument.  Localestring is read destructively. A warning is
   given if a non-well-formed rule is encountered. At each call, the
   next well-formed rule is found and the left-hand-side and
   right-hand-side are returned in *lhs_p and *rhs_p. If no more rule
   is found, return two NULL strings. Localereader uses the library
   function strtok, and will not work if interleaved with another
   function that calls strtok. */

void localereader(char *localestring, char **lhs_p, char **rhs_p) {
  char *p;

  p = strtok(localestring,":");
  while (p!=NULL) {
    char* tmp = strchr(p,'=');
    if (tmp==NULL) {
      fprintf(stderr, 
	      "%s: Warning: ignoring invalid locale table entry \'%s\'.\n",
	      MYNAME, p);
      p = strtok(NULL,":");
    } else {
      *tmp='\0';
      *lhs_p = p;
      *rhs_p = tmp+1;
      return;
    }
  }
  *lhs_p = NULL;
  *rhs_p = NULL;
  return;
}

/* ---------------------------------------------------------------------- */
/** wrap calloc with out of memory message */
void *safecalloc(size_t nmemb, size_t size) {
  void *result = calloc(nmemb, size);
  if (result==NULL) {
    fprintf(stderr, "%s: Out of Memory\n", MYNAME);
    exit(1);
  }
  else 
    return result;
}

/* ---------------------------------------------------------------------- */
/** create allocated string by appending given strings */
char *strappend(char *s1, char *s2) {
  int s1len, s2len;
  char *result;

  s1len = strlen(s1);
  s2len = strlen(s2);

  result = (char *)safecalloc(s1len+s2len+1,sizeof(char));
  strcpy(result,s1);
  strcpy(result+s1len,s2);
  
  return result;
}

/* ---------------------------------------------------------------------- */
/** make allocated clone of string */

char *strclone(char *s) {
  return strappend(s, "");
}

/* ---------------------------------------------------------------------- */
/** destructively strip s of non-dialchars. Returns s. */

char *strip(char *s) {
  char *from = s;
  char *to = s;

  while (*from!='\0') {
    if (is_dialchar(*from)) {
      *to = *from;
      to++;
    }
    from++;
  }
  *to = '\0';
  return s;
}  


/* ---------------------------------------------------------------------- */
/** localphone: the main procedure of this program. Replaces the
    longest prefix of the phone number inphone that matches a rule in
    the global variable localestring */

char *localphone(char *inphone) {
  int matchlen;
  char *best_lhs, *best_rhs, *outphone;
  char *lhs, *rhs; 

  matchlen = -1;

  /* try rules (lhs/rhs) one by one - use the longest match. */

  localereader(localestring,&lhs,&rhs);
  while (lhs!=NULL) {
    int this_matchlen = match_prefix(lhs,inphone);
    if (this_matchlen > matchlen) {  /* found better than previous match */
      matchlen = this_matchlen;
      best_rhs = rhs;
      best_lhs = lhs;
    }
    localereader(NULL,&lhs,&rhs);
  }
  
  if (matchlen < 0) {

    /* did not find a match */
    fprintf(stderr, "%s: Warning: no matching rule in locale for %s. "
	    "Not replacing.\n", MYNAME, inphone);
    outphone = inphone;

  } else {

    /* remove leading '-' from unmatched part of phone number */
    if (inphone[matchlen]=='-' && !is_dialchar('-'))
      matchlen++;  
    
    /* do the replacement */
    outphone = strappend(best_rhs,inphone+matchlen);
    
    if (verbose)
      fprintf(stderr, "Replacement phone by rule \'%s\' -> \'%s\': %s\n", 
	      best_lhs, best_rhs, outphone);
  }

  return outphone;
}

/* ---------------------------------------------------------------------- */
/** get a line from file. Line can have at most LINESIZE characters.
  Returns NULL when no more line in file. Line is read destructively,
  thus the caller must make a copy before calling getline again */

char *getline(FILE *fin) {
  static char acc[LINESIZE];
  char *p;

  p = fgets(acc,LINESIZE,fin);
  if (p==NULL)
    return NULL;

  /* fgets is supposed to remove the trailing newline, but on my
     system it does not. Thus, we remove it here. */

  p=strchr(acc,'\n');
  if (p!=NULL) {
    *p='\0';
  }
  return acc;
}

/* ---------------------------------------------------------------------- */
/** finds the first line that starts with key+whitespace in
    file. returns NULL if line is not found, or if file cannot be
    read.  Ignores lines that start with "#". Beginning and trailing
    whitespace is ignored and stripped. */

char *getkeyf(char *key, char *file) {
  char *line, *endline;
  FILE *fin;
  int n = strlen(key);

  fin = fopen(file, "r");
  if (fin==NULL)
    return NULL;

  while ((line = getline(fin)) != NULL) {
    while (isblank(*line))
      line++;
    if (*line=='#' || strncmp(key, line, n)!=0)
      continue;
    line += n;
    if (!(isblank(*line) || *line==0))
      continue;
    while (isblank(*line))
      line++;
    /* strip trailing whitespace */
    endline = line+strlen(line)-1;
    while (endline>=line && isblank(*endline)) {
      *endline = 0;
      endline--;
    }
    fclose(fin);
    return strclone(line);
  }
  
  fclose(fin);
  return NULL;
}
  
/* ---------------------------------------------------------------------- */
/** prints summary of command line options. */

usage(FILE *fout) {
  fprintf(fout, "\
"MYNAME" v"VERSION". Figure out the local way of dialing a phone number.

Usage: "MYNAME" [options] phone-number
       "MYNAME" [options] alias
       "MYNAME" [options] -

Options and arguments: 
         -h, --help      help
         -v              verbose
         -n              suppress newline
         -l locale       set locale string
         -f localename   set locale by name
         -d dialchars    set valid digits (default \""DIALCHARS"\")
         -s              strip output of non-dialchars
         phone-number    a fully qualified phone number (starts with \"+\")
         alias           a nickname for a phone number
         -               run as a filter
");
}

/* ---------------------------------------------------------------------- */
/* reads the appropriate information from the command line, the
   environment, and stdin, then calls localphone, then writes the
   output to stdout. */

main(int ac, char *av[]) {
  /* command line parameters */
  int nonewline = 0;           /* suppress newline at end of output? */
  int filter = 0;              /* run as filter? */
  char *inphone = NULL;        /* input phone number */
  char *nicklocale = NULL;     /* nickname for locale name */
  char *nickphone = NULL;       /* nickname for phone number */

  /* other local variables */
  char *p;                     /* generic temporary */
  char *outphone;              /* output phone number */
  char *home = getenv("HOME"); /* the user's home directory */

  /* read command line */

  ++av, --ac;
  while (ac > 0) {
    if (!strcmp(*av, "-h") || !strcmp(*av, "--help")) {     /* help */
      usage(stdout); exit(0);
    } else if (!strcmp(*av, "-v")) { /* verbose */
      verbose++;
    } else if (!strcmp(*av, "-n")) { /* no newline */
      nonewline++;
    } else if (!strcmp(*av, "-l")) { /* locale */
      if (ac<=1)
	goto missing_arg;
      ++av, --ac;
      localestring = *av;
    } else if (!strcmp(*av, "-d")) { /* digit characters */
      if (ac<=1)
	goto missing_arg;
      ++av, --ac;
      dialchars = *av;
    } else if (!strcmp(*av, "-f")) { /* from: specify a locale name */
      if (ac<=1)
	goto missing_arg;
      ++av, --ac;
      nicklocale = *av;
    } else if (!strcmp(*av, "-s")) { /* s: strip output */
      stripflag++;
    } else if (!strcmp(*av, "-")) { /* "-" option: run as filter */
      filter=1;
      inphone = NULL;
      nickphone = NULL;
    } else if (*av[0]=='-') { /* invalid option */
      fprintf(stderr, ""MYNAME": illegal option %s. Try --help for more info.\n", *av);
      exit(1);
    } else if (*av[0]=='+') { /* starts with "+": phone number */
      inphone = *av;
      nickphone = NULL;
      filter = 0;
    } else { /* not an option, does not start with "+": must be an alias */
      nickphone = *av;
      filter=0;       /* if - was also given, ignore */
      inphone=NULL;   /* if other phone number was given, ignore */
    }
    av++, ac--;
  }
  goto done;
  
 missing_arg:        /* an option was missing a required argument */

  fprintf(stderr, ""MYNAME": option %s needs an argument. "
	  "Try --help for more info.\n", *av);
  exit(1);

 done:               /* done with reading options; syntax seemed okay */

  /* ---------------------------------------------------------------------- */
  /* determine inphone */

  if (inphone==NULL && !filter && nickphone==NULL) {
    fprintf(stderr, 
	    ""MYNAME": no phone number specified. Try --help for more info.\n");
    exit(1);
  }

  /* if "-" specified on command line, read phone number from stdin */
  if (inphone==NULL && filter) {
    inphone = strclone(getline(stdin));
    if (inphone == NULL) {
      fprintf(stderr,
              "%s: could not read from stdin.\n",
              MYNAME);
      exit(1);
    }
    /* remove trailing newline */
    p=strchr(inphone,'\n');
    if (p!=NULL) {
      *p='\0';
    }
  } else if (inphone==NULL && nickphone != NULL) { /* read from files */
    if (home!=NULL) {
      char *useralias = strappend(home, "/" USERALIAS);
      inphone = getkeyf(nickphone, useralias);
    }
    if (inphone == NULL)
      inphone = getkeyf(nickphone, SYSTEMALIAS);
    if (inphone == NULL) {
      fprintf(stderr, 
	      "%s: alias %s not found.\n", 
	      MYNAME, nickphone);
      exit(1);
    }      
  }

  if (verbose)
    fprintf(stderr, "Phone: %s\n", inphone);

  /* ---------------------------------------------------------------------- */
  /* determine localestring */

  if (localestring!=NULL && nicklocale!=NULL) {
    fprintf(stderr, "%s: options -l and -f incompatible.\n", MYNAME);
    usage(stderr); exit(1);
  }

  if (nicklocale==NULL) 
    nicklocale="default";

  /* if not specified on command line, read from environment */
  if (localestring == NULL) {
    if (home!=NULL) {
      char *userlocale = strappend(home, "/" USERLOCALE);
      localestring = getkeyf(nicklocale, userlocale);
    }
    if (localestring == NULL)
      localestring = getkeyf(nicklocale, SYSTEMLOCALE);
    if (localestring == NULL) {
      fprintf(stderr, 
	      "%s: locale %s not found.\n", 
	      MYNAME, nicklocale);
      exit(1);
    }
  }
  
  if (verbose)
    fprintf(stderr, "Locale: \'%s\'\n", localestring);

  /* ---------------------------------------------------------------------- */
  /* transform inphone according to localestring */

  outphone = localphone(inphone);

  /* ---------------------------------------------------------------------- */
  
  if (stripflag) {
    outphone = strip(outphone);
    if (verbose)
      fprintf(stderr, "Stripped: %s\n", outphone);
  }

  /* ---------------------------------------------------------------------- */
  /* print the output and quit */

  printf("%s", outphone);
  if (!nonewline)
    printf("\n");

  return 0;
}
